(* Compiler.ml for Nifty50 by Aamir Jahan, COMS W4115 *)

open Ast
module StringMap = Map.Make(String)

(*stores global variables *)
type varSymbolTableEntry = {
  data_type : datatypes;
}

(*stores functions *)
type funcSymbolTableEntry = {
  ftype : datatypes;
  fparameters : datatypes StringMap.t;
  flocals :  datatypes StringMap.t;
  fbodylist : stmt list;
}

(*stores Interrupt handlers and local variables*)
type interruptSymbolTableEntry = {
  inlocals : datatypes StringMap.t;
  inbodylist : stmt list;
}

(*stores user-defined types *)
type typeTableEntry = {
  thetype : datatypes;
  properties : datatypes StringMap.t;
  functions : funcSymbolTableEntry StringMap.t;
}

(*environment that describes all types*)
type environment = {
  niftyVarSymbolTable : datatypes StringMap.t;
  niftyFuncSymbolTable : funcSymbolTableEntry StringMap.t;
  niftyInterruptSymbolTable : interruptSymbolTableEntry StringMap.t;
  niftyTypesSymbolTable : typeTableEntry StringMap.t;
}

let checkFunctionBody theBody env construct = 

let rec output_function_expr = function
    Sliteral(l) -> l
  | Fliteral(l) -> string_of_float l
  | Iliteral(l) ->  string_of_int l
	| Pliteral(l) -> l
  | Id(s) ->  s 
  | Binop(e1, o, e2) -> "(" ^
      string_of_expr e1 ^ " " ^
      (match o with
  			Add -> "+" 
			| Sub -> "-" | Mult -> "*" | Div -> "/"
      | Equal -> "==" | Neq -> "!="
      | Less -> "<" | Leq -> "<=" | Greater -> ">" | Geq -> ">=" 
      | Or -> "or"      | And -> "and" 
      | Bsr -> ">>" | Bsl -> "<<" | Xor -> "xor" ) ^ " " ^
      string_of_expr e2 ^ ")"
  | Unop(o, e)-> "(" ^ 
    (match o with
      Not -> "!" 
      ) ^ output_function_expr e^")" 
  | Assign(v, e) -> output_function_expr v ^ " = " ^ output_function_expr e
  | Call(f, el) -> 
  (match f with
    "print" -> "printf" ^ "(" ^ String.concat ", &" (List.map output_function_expr el) ^ ")"
    | _ -> f ^ "(" ^ String.concat ", &" (List.map output_function_expr el) ^ ")"
  ) 
  | MethodCall(v, f, el)->
  (match f with
     _         -> v ^ "." ^ f ^ "(" ^ String.concat ", " (List.map output_function_expr el) ^ ")"
  )
  | Noexpr -> ""
  | CastType(a, b)->""
  | Address(a, b)-> "Util_Move((void *)(&"^ a ^ "), (void*)(&" ^ output_function_expr b ^ "), sizeof(" ^ a ^ "))";  
  | GetAddress(a)-> "(&" ^ a ^")"
  | ArrayIndex(a, b)-> a ^ "[" ^  b ^ "]"
  | Method(a, b)-> output_function_expr a ^ "." ^ output_function_expr b
  | Signal(a, b)-> "Interrupt_SetISR(" ^ a ^ "____Handler, &"^ a ^");\n Interrupt_SetSignal(" 
          ^ output_function_expr b ^ ", &" ^ a ^ ")"
  | Swap(a)-> "Util_Swap((void *)&" ^  a ^ ", sizeof(" ^  a ^ "))"
  | Map(a, b, c, d)->
      "Array_Map((void *)" ^a^", (void (*)(void *))"^b^", "^output_function_expr c^", " ^ output_function_expr d ^")"
  | Variable(a)-> "" in

let rec output_function_stmts = function
  | Block(stmts)  -> "{\n" ^ String.concat "" (List.map output_function_stmts stmts) ^ "}\n";
  | Expr(expr)    -> output_function_expr expr ^ ";\n";
  | Return(expr)  -> "return " ^ output_function_expr expr ^ ";\n"
  | If(e, s, Block([])) -> "if (" ^ output_function_expr e ^ ")\n" ^ output_function_stmts s;
  | If(e, s1, s2)       -> "if (" ^ output_function_expr e ^ ")\n" ^
      output_function_stmts s1 ^ "else\n" ^ output_function_stmts s2;
  | For(e1, e2, e3, s)  -> "for (" ^ output_function_expr e1  ^ " ; " ^ output_function_expr e2 ^ " ; " ^
      string_of_expr e3  ^ ") " ^ String.concat "" (List.map output_function_stmts s) ;
  | While(e, s)         -> "while (" ^ string_of_expr e ^ ")\n{" ^ String.concat "" (List.map output_function_stmts s) ^"}"


  in List.map output_function_stmts (List.rev theBody);; 

(* prints the type  *)
let rec output_type key character = function
  | Port                -> "unsigned char " ^ character ^ key
  | Int                 -> "int " ^ character ^ key
  | Double              -> "double " ^ character ^ key
  | String              -> "char *" ^ character ^ key
  | NewType(newtype)    -> newtype ^ character ^ " " ^ key
  | Array(basetype, sz) -> (output_type key character basetype ) ^ "[" ^ sz ^ "]" 

let rec output_typefunc key = function
  Void                  -> "void " ^ " " ^ key
  | Port                -> "unsigned char "  ^ key
  | Int                 -> "int " ^ key
  | Double              -> "double " ^ key
  | String              -> "char *" ^ key
  | NewType(newtype)    -> newtype ^ " " ^ key

let output_newtype = function
| NewType(newtype) -> newtype ^ " "

(* returns the global variable, AST checked post semantically*)
 let output_globals global_table env =  
  let output_each_global key value =
    print_endline (output_type key "" value  ^ ";") in
  StringMap.iter output_each_global global_table

 let output_params param_table =  
  let output_each_param key value =
    print_string (output_type key "*" value ^ ", ") in
  StringMap.iter output_each_param param_table


(* spits out the interrupt handlers in C form, post semantically checked AST *)
let output_interrupts interrupt_table env = 
  let output_each_interrupt key value =
    print_endline ("Interrupt " ^ key ^ ";"); in
    StringMap.iter output_each_interrupt interrupt_table;
  let output_each_interruptbody key value =
    let interruptbody = 
    "void " ^ key ^ "____Handler (int sig)\n{" in
    print_endline (interruptbody);
    output_globals value.inlocals env;
    (*eventually output the body here as well*)
    List.map print_endline (checkFunctionBody value.inbodylist env Interrupt); 
    (* List.map print_endline (List.map output_function_stmts value.inbodylist); the gangsta line*)
    print_endline ("}\n") in
    StringMap.iter output_each_interruptbody interrupt_table;;


 let output_functions function_table typeFunction env = 
  let output_each_function key value =
    print_string (output_typefunc key value.ftype ^ "("); 
    output_params value.fparameters; 
    if typeFunction != Ast.Void then print_string ((output_newtype typeFunction) ^" *itself");
    print_endline (")");
    print_endline ("{"); 
    output_globals value.flocals env;
    (*eventually output the body here as well*)
    List.map print_endline (checkFunctionBody value.fbodylist env Function); 
    (* List.map print_endline (List.map output_function_stmts value.fbodylist); the gangsta line*)
    print_endline ("}\n") in
    StringMap.iter output_each_function function_table;;

 let output_types types_table env = 
  let output_each_type key value =
    print_endline ("typedef struct {"); 

    output_globals value.properties env;
    print_endline ("} " ^ key ^ ";\n"); 
    (*eventually output the body here as well*)
    output_functions value.functions value.thetype env in
    StringMap.iter output_each_type types_table;;

(*utility to print out keys of each string Map*)
let print_vars key value =
    print_string(key ^ " \n");
    (*array type checking.  This is amazing!! *)
    let thetype = Ast.Array(Int, "function") and thattype = value in 
    if thetype = thattype then print_endline "Yes" 
  else print_endline "No";;

let print_funcs key value = 
    print_string("Function "^ key ^": \n");
    print_endline "Parameters:";
    StringMap.iter print_vars value.fparameters;
    print_endline "Local variables:";
    StringMap.iter print_vars value.flocals;;

let print_interrupts key value = 
    print_string("Interrupt "^ key ^": \n");
    print_endline "Local variables:";
    StringMap.iter print_vars value.inlocals;;

let print_types key value = 
    print_string("Nifty50 Type "^ key ^": \n");
    print_endline "Properties:";
    StringMap.iter print_vars value.properties;
    print_endline "Methods:";
    StringMap.iter print_funcs value.functions;;    

(*creates symbol tables for variables *)
let createNiftyVarSymbolTable map (var_elem: var_decl list) = 
  List.fold_left 
  (fun map thelist -> 
    if StringMap.mem thelist.vname map 
    then 
      raise(Failure("Compiler error: variable named \"" ^ thelist.vname ^ "\" already exists.")) 
    else 
      StringMap.add thelist.vname thelist.vtype map) 
  map var_elem

let createNiftyFuncSymbol (var_elem: func_decl) = 
  {
    ftype = var_elem.rtype;
    fparameters = createNiftyVarSymbolTable StringMap.empty var_elem.formals;
    flocals = createNiftyVarSymbolTable StringMap.empty var_elem.locals;
    fbodylist = var_elem.body;
  }
    

(*creates symbol table for functions *)
let createNiftyFuncSymbolTable map (var_elem: func_decl list) = 
  List.fold_left (fun map thelist -> 
    if StringMap.mem thelist.fname map then raise(Failure("Compiler error: function named \'" ^ thelist.fname ^ "\' already exists.")) else 
    StringMap.add thelist.fname (createNiftyFuncSymbol thelist) map) map var_elem

(*creates symbol table for interrupts *)
let createNiftyInterruptSymbol (var_elem: interrupt_decl) = 
  {
    inlocals = createNiftyVarSymbolTable StringMap.empty var_elem.ilocals;
    inbodylist = var_elem.ibody;
  }
  
(*creates symbol table for interrupts *)
let createNiftyInterruptSymbolTable map (var_elem: interrupt_decl list) = 
  List.fold_left (fun map thelist -> 
    if StringMap.mem thelist.iname map then raise(Failure("Compiler error:Interrupt handler named \'" ^ thelist.iname ^ "\' already exists.")) else 
    StringMap.add thelist.iname (createNiftyInterruptSymbol thelist) map) map var_elem

(*creates symbol table for new types *)
let createNiftyTypesSymbol (var_elem: type_decl) = 
  {
    thetype = var_elem.ytype;
    properties = createNiftyVarSymbolTable StringMap.empty var_elem.yproperties;
    functions = createNiftyFuncSymbolTable StringMap.empty var_elem.yfunctions;
  }

(*creates symbol table for new types*)
let createNiftyTypesSymbolTable map (var_elem: type_decl list) = 
  List.fold_left (fun map thelist -> 
    if StringMap.mem thelist.yname map then raise(Failure("Compiler error: Type named \'" ^ thelist.yname ^ "\' already exists.")) else 
    StringMap.add thelist.yname (createNiftyTypesSymbol thelist) map) map var_elem


let header =
"#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include \"Util.h\"
#include \"Interrupt.h\"
#include \"Array.h\"
"

(* Translates program in AST form into bytecode program.*)
(* Includes exception handling. *)
		let translate (globals, functions, interrupts, types) =

    let var_table = createNiftyVarSymbolTable StringMap.empty globals in

    let fun_table = createNiftyFuncSymbolTable StringMap.empty functions in

    let interrupt_table = createNiftyInterruptSymbolTable StringMap.empty interrupts in

    let type_table = createNiftyTypesSymbolTable StringMap.empty types in
    (* StringMap.iter print_types type_table; *)
    let env = 
    {
      niftyVarSymbolTable = var_table;
      niftyFuncSymbolTable = fun_table;
      niftyInterruptSymbolTable = interrupt_table;
      niftyTypesSymbolTable = type_table;

    } in

    print_endline header; 
    output_globals var_table env;
    output_types type_table env;
    output_interrupts interrupt_table env;
    output_functions fun_table Void env;

(*End of File*)